# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from bawr import utils
from pathlib import Path 
import hashlib

class RegexReplacePreprocessor:

    def __init__(self, translations, ignore_case = True):
        self.translations = translations
        self.ignore_case = ignore_case
        self.processor = utils.StringReplacer(translations, ignore_case)

    def __call__(self, env, file, iconset):
        name = hashlib.sha256("{}{}{}".format(self.__class__.__name__, iconset, file.stem).encode()).hexdigest()
        cache_dir = utils.get_cache_dir(env)
        path = Path(cache_dir, "{}.svg".format(name))
        if path.exists() and path.stat().st_mtime > file.stat().st_mtime:
            return path
        with open(file, mode='r') as f:
            txt = f.read()
            txt = self.processor(txt)
            if self.processor.changed:
                with open(path, mode="w") as w:
                    w.write(self.processor(txt))
                    return path
        return file

    def __repr__(self):
        return "{}"




# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from pathlib import Path
from bawr.glyph import Glyph
from bawr.tool_inkscape import InkscapeTool
from bawr import utils
import os

try:
    from PIL import Image
except:
    Image = None

try:
    from rectpack.packer import SORT_AREA, PackingBin, PackingMode, MaxRectsBssf
    from rectpack import newPacker
except:
    newPacker = None


def pack(cells, width, height):
    packer = newPacker(
        mode=PackingMode.Offline, 
        bin_algo=PackingBin.BBF, 
        pack_algo=MaxRectsBssf, 
        sort_algo=SORT_AREA, 
        rotation=False
    )
    for i, cell in enumerate(cells):
        packer.add_rect(cell.width, cell.height, rid=i)
    packer.add_bin(width, height)
    packer.pack()
    for bin in packer:
        for rect in bin:
            cells[rect.rid].x = rect.left
            cells[rect.rid].y = rect.top - rect.height


class Cell:
    def __init__(self, file, name, size, margin):
        self.file = file
        self.name = name
        self.width = size
        self.height = size
        self.x = 0
        self.y = 0
        self.margin = margin


class Atlas:

    # Config
    name = 'atlas'
    collections = []
    width = 512
    sizes = (16, 32, 64)

    # Generated
    height = 0
    icons = []
    cells = []
    instance = None
    output = None

    def build(self, env):
        print("[Atlas Texture] Start ...")

        self.icons = []
        self.cells = []

        if not env.INKSCAPE_PATH:
            print("[Atlas Texture] Cancelled: Inkscape not available.")
            return

        if not Image:
            print("[Atlas Texture] Cancelled: Pillow not available.")
            return

        if not newPacker:
            print("[Atlas Texture] Cancelled: Reckpack not available")
            return

        code = 0
        for icon_set_cls in utils.as_iterable(self.collections):
            icon_set = icon_set_cls()
            if icon_set.select:
                for file_name, glyph_name in icon_set.select:
                    self.icons.append(Glyph(code, glyph_name, Path(env.BAWR_SOURCE_DIR, icon_set.src, file_name + ".svg"), icon_set_cls))
                    code += 1
            else:
                source = Path(env.BAWR_SOURCE_DIR, icon_set.src)
                if source.exists():
                    for file in sorted(os.listdir(source)):
                        if file[-4:].lower() != '.svg':
                            continue
                        stem = file[:-4].lower()
                        self.icons.append(Glyph(code, stem, Path(env.BAWR_SOURCE_DIR, icon_set.src, file), icon_set_cls))
                        code += 1
                else:
                    print("[Atlas Texture] Invalid source dir: {}".format(source.absolute()))
        self.generate_individual_cells(env)
        self.generate_texture(env)
        print("[Atlas Texture] {}".format(self.output))


    def generate_individual_cells(self, env):     
        cache_dir = utils.get_cache_dir(env) 
        inkscape = InkscapeTool(env)          
        for glyph in self.icons:
            file = glyph.path
            margin = 0

            if glyph.source.options:
                processors = glyph.source.options.get('atlas_preprocessors', [])
                margin = glyph.source.options.get('atlas_margin', 0)
                for p in processors:
                    file = p(env, file, glyph.source.__name__)
            
            if margin > 1:
                margin = 1.0 / margin
            elif margin == 1:
                margin = 0
            elif margin < 0:
                margin = 0
            
            for size in self.sizes:
                png = Path(cache_dir, "{}_{}_{}.png".format(glyph.source.__name__, file.stem, size))
                if (not png.exists()) or png.stat().st_mtime < file.stat().st_mtime:
                    inkscape(file, png, size, int(margin * size))
                self.cells.append(Cell(png, glyph.name, size, int(margin * size)))

        pack(self.cells, self.width, self.width * 20)


    def generate_texture(self, env):
        height = 0
        for cell in self.cells:
            h = cell.y + cell.height
            if h > height:
                height = h
        atlas = Image.new("RGBA", (self.width, height))
        for cell in self.cells:
            self.apply_image(cell, cell.margin, atlas)
        self.output = Path(env.BAWR_OUTPUT_DIR, (self.name or 'atlas') + '.png')
        atlas.save(self.output, quality=99, optimize=True)
        self.height = height


    def apply_image(self, cell, margin, atlas):
        icon = Image.open(cell.file)
        pos = (cell.x, cell.y)
        if margin > 0:
            pos = (pos[0] + margin, pos[1] + margin) 
        atlas.paste(icon, pos)





# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import subprocess
import sys
import shlex

class InkscapeTool:

    def __init__(self, env):
        self.env = env

    def __call__(self, svg_file, png_file, size, margin=0):
        if margin > 0:
            size -= 2*margin

        cmd_exec =  [
            str(self.env.INKSCAPE_PATH),
            '--export-background-opacity=0',
            f'--export-width={size}',
            '--export-type=png',
            '-o', str(png_file), 
            str(svg_file)
        ]
        print(f"[EXEC] {shlex.join(cmd_exec)}")
        process = subprocess.Popen(cmd_exec, shell=False)
        try:
            err = process.wait(30)
            if (err):
                sys.exit(f"Error: Inkscape terminated with status = {err}")
        except subprocess.TimeoutExpired:
            sys.exit(f"Error: Inkscape timeout")



# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import argparse
from bawr import config_parser as cp

VERSION = '0.0.9'

def banner():
    print(r'''
  ______  ___  _    _______ 
  | ___ \/ _ \| |  | | ___ \
  | |_/ / /_\ \ |  | | |_/ /
  | ___ \  _  | |/\| |    / 
  | |_/ / | | \  /\  / |\ \ 
  \____/\_| |_/\/  \/\_| \_|

  version {}                          
  (c) Copyright 2021 Frank David Martinez Muñoz  
      Licensed under MIT License.

    '''.format(VERSION))


def main(args):
    banner()
    cfg = cp.Parser(args.cfg, dict(src_dir=args.src, out_dir=args.out))


def command_line_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--out", help="Output directory path", default="build")
    parser.add_argument("--src", help="Source directory path", default=".")
    parser.add_argument("--cfg", help="Config file", default="config.py")
    return parser.parse_args()


if __name__ == '__main__':
    main(command_line_arguments())



# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from bawr.iconset import IconSet
from bawr.gen_atlas import Atlas
from bawr.gen_cpp_embedded import CppEmbedded
from bawr.gen_cpp_font_header import CppFontHeader
from bawr.gen_font import Font
from bawr.gen_imgui_loader import ImGuiFontLoader
from bawr.environment import Environment
from bawr.preprocessor import RegexReplacePreprocessor
from bawr.gen_cpp_atlas_header import CppAtlasHeader













# Copyright 2021 Frank David Martinez Muñoz
# Copyright 2024 Lutz Schönemann
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from pathlib import Path
import os
from bawr.glyph import Glyph
from bawr import utils
from bawr.tool_fontforge import FontForgeTool

class Font:

    # Config
    copyright = "Copyright 2021 Frank David Martinez M."
    name = "icons"
    family = "icons"
    version = ''
    start_code = 0xe000
    collections = []
    transformation = []
    output_formats = ('ttf',)
    verbose = False

    # Generated
    icons = []
    instance = None

    def build(self, env):
        print("[Font Forge] Start ...")
        code = self.start_code
        for icon_set_cls in utils.as_iterable(self.collections):
            icon_set = icon_set_cls()
            if icon_set.select:
                for file_name, glyph_name in icon_set.select:
                    self.icons.append(Glyph(code, glyph_name, Path(env.BAWR_SOURCE_DIR, icon_set.src, file_name + ".svg"), icon_set_cls))
                    code += 1
            else:
                source = Path(env.BAWR_SOURCE_DIR, icon_set.src)
                if source.exists():
                    for file in sorted(os.listdir(source)):
                        if file[-4:].lower() != '.svg':
                            continue
                        stem = file[:-4].lower()
                        self.icons.append(Glyph(code, stem, Path(env.BAWR_SOURCE_DIR, icon_set.src, file), icon_set_cls))
                        code += 1
                else:
                    print("[error] Invalid source dir: {}".format(source.absolute()))
        self.end_code = code - 1 if code != self.start_code else None
        self.generate(env)
        self.run(env)

        print("")
        for fmt in self.output_formats:
            print("[Font Forge] {}".format(Path(env.BAWR_OUTPUT_DIR, "{}.{}".format(self.name, fmt))))


    def generate(self, env):
        cache_dir = utils.get_cache_dir(env) 
        with open(Path(Path(__file__).parent, 'gen_font_ff_template.py')) as f:
            pre, discard, pos = f.read().partition("#@generated")
            with open(Path(cache_dir, '_ff_.py'), 'w') as out:
                out.write(pre)
                out.write("class config:\n")
                out.write("  font_copyright = '{}'\n".format(self.copyright))
                out.write("  font_family = '{}'\n".format(self.family))
                out.write("  font_name = '{}'\n".format(self.name))
                out.write("  font_version = '{}'\n".format(self.version))
                out.write("\n")

                for icon_set_cls in utils.as_iterable(self.collections):
                    out.write("{}_opt = {}\n".format(icon_set_cls.__name__, icon_set_cls.options))
                out.write("\n")

                out.write("global_transform = {}\n".format(self.transformation or None))
                out.write("\n")

                out.write("output_formats = {}\n".format(self.output_formats or ('ttf',)))
                out.write("\n")

                out.write("verbose = {}\n".format(self.verbose))
                out.write("\n")

                out.write("select = [\n")
                for glyph in self.icons:
                    out.write("  ({}, '{}', r'{}', {}_opt),\n".format(hex(glyph.code), glyph.name, glyph.path.absolute(), glyph.source.__name__))
                out.write("]\n")
                out.write("\n")

                out.write("build_dir = r'{}'".format(env.BAWR_OUTPUT_DIR.absolute()))
                out.write(pos)

    def run(self, env):
        cache_dir = utils.get_cache_dir(env) 
        tool = FontForgeTool(env)
        tool(Path(cache_dir, '_ff_.py'))



# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from pathlib import Path
from datetime import datetime

class CppAtlasHeader:
    name = None
    namespace = None
    constexpr = True
    source = None
    instance = None
    output = None

    def build(self, env):
        print(f"[Cpp Atlas Header] Start ...")
        if not self.source or not self.source.instance:
            print(f"[Cpp Atlas Header] Invalid source (CppAtlasHeader.source)")
            return
        
        atlas = self.source.instance
        hpp_file = Path(env.BAWR_OUTPUT_DIR, (self.name or atlas.name or 'icons') + '_cells.hpp')
        print(f"[Cpp Atlas Header] {hpp_file}")

        constexpr = 'constexpr' if self.constexpr else 'const'

        with open(hpp_file, 'w') as f:
            f.write("#pragma once\n\n")
            f.write("struct ImVec2;\n")
            f.write(f"namespace {self.namespace or 'icons'} {{\n")
            f.write( " struct uv { float x; float y; inline operator ImVec2 const& () const { return *reinterpret_cast<const ImVec2*>(this); } };\n")
            f.write( " struct frame { uv uv0; uv uv1; };\n")

            icons = atlas.cells
            width = atlas.width
            height = atlas.height
            for size in atlas.sizes:
                f.write(f" struct sz{size} {{\n")
                for icon in icons:
                    u0 = icon.x / width
                    v0 = icon.y / height
                    u1 = u0 + (icon.width / width)
                    v1 = v0 + (icon.height / height)
                    if icon.width == size:
                        f.write(f"   static {constexpr} frame {icon.name.ljust(24)} {{{{{u0:.10f},{v0:.10f}}},{{{u1:.10f},{v1:.10f}}}}};\n")
                f.write(f" }};\n")
            f.write( "}\n")

        self.output = hpp_file


# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from pathlib import Path

class ImGuiFontLoader:

    # Config
    font = None
    header = None
    data = None
    name = None
    namespace = 'icons'

    # Generated
    instance = None
    output = None

    def build(self, env):
        print("[ImGui Font Loader] Start ...")

        if self.font is None or self.font.instance is None:
            print(f"[ImGui Font Loader] Invalid Font (ImGuiFontLoader.font)")
            return
        if self.data is None or self.data.instance is None:
            print(f"[ImGui Font Loader] Invalid Data (ImGuiFontLoader.data)")
            return
        if self.header is None or self.header.instance is None:
            print(f"[ImGui Font Loader] Invalid Header (ImGuiFontLoader.header)")
            return

        instance = self.data.instance
        build_dir = env.BAWR_OUTPUT_DIR
        namespace = self.namespace or self.instance.namespace

        header_file = Path(build_dir, self.name or instance.name + "_loader.hpp")
        print(f"[ImGui Font Loader] {str(header_file)}")

        data_header_file = instance.output_header        
        with open(header_file, 'w') as f:
            f.write( "#pragma once\n\n")
            f.write(f'#include <imgui.h>\n')
            f.write(f'#include "{data_header_file.name}"\n')
            f.write(f'#include "{self.header.instance.output.name}"\n\n')
            f.write(f"namespace {namespace} {{\n")
            f.write( " namespace Font {\n")
            f.write(f"  static const ImWchar ranges[] = {{ {self.header.instance.namespace}::Font_StartCode , {self.header.instance.namespace}::Font_EndCode, 0 }};\n")
            f.write( "  inline ImFont* Load(ImGuiIO& io, const float size, ImFontConfig* config) {\n")
            f.write(f"   void* data = const_cast<unsigned int*>({self.header.instance.namespace}::data::DATA);\n")
            f.write( "   if (config) {\n")
            f.write( "    config->FontDataOwnedByAtlas = false;\n")
            f.write(f"    return io.Fonts->AddFontFromMemoryTTF(data, {self.header.instance.namespace}::data::SIZE, size, config, ranges);\n")
            f.write( "   }\n")
            f.write( "   else {\n")
            f.write( "    ImFontConfig dconf;\n")
            f.write( "    dconf.FontDataOwnedByAtlas = false;\n")
            f.write(f"    return io.Fonts->AddFontFromMemoryTTF(data, {self.header.instance.namespace}::data::SIZE, size, &dconf, ranges);\n")
            f.write( "   }\n")
            f.write( "  }\n")
            f.write( " }\n")
            f.write( "}\n")

        self.output = header_file


# Copyright 2021 Frank David Martinez M.
# Copyright 2024 Lutz Schönemann
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import fontforge, psMat
import os, sys, math

# +-------------------------------------------------------------------+
# | Start Generated Glyphs                                            |
# | defines: config, select, options, build_dir, global_transform,    |
# |          output_formats, verbose                                  |

#@generated

# | End Generated Glyphs                                              |
# +-------------------------------------------------------------------+

try:
    if not os.path.exists(build_dir):
        os.mkdir(build_dir)
except:
    sys.exit("Error creating directory: %s" % build_dir)   


def parse_transformation(t):
    '''
        ('t|translate', x, y)
        ('s|scale', x, y)
        ('r|rotate', deg)
        ('x|skew', deg)
    '''
    m = None
    if t:
        m = psMat.identity()
        for tx in t:
            if tx[0] == 't' or tx[0] == 'translate':
                m = psMat.compose(m, psMat.translate(tx[1], tx[2]))
            elif tx[0] == 'r' or tx[0] == 'rotate':
                m = psMat.compose(m, psMat.rotate(math.radians(tx[1])))
            elif tx[0] == 's' or tx[0] == 'scale':
                m = psMat.compose(m, psMat.scale(tx[1], tx[2]))
            elif tx[0] == 'x' or tx[0] == 'skew':
                m = psMat.compose(m, psMat.skew(math.radians(tx[1])))
    return m


font = fontforge.font()
font.encoding = "UnicodeFull"
font.copyright = config.font_copyright
font.familyname = config.font_family
font.fontname = config.font_name
font.fullname = config.font_name
font.version = config.font_version.strip() or font.version

# add latin characters to be used in
ligaChars = [
      (0x30, 'zero'),
      (0x31, 'one'),
      (0x32, 'two'),
      (0x33, 'three'),
      (0x34, 'four'),
      (0x35, 'five'),
      (0x36, 'six'),
      (0x37, 'seven'),
      (0x38, 'eight'),
      (0x39, 'nine'),

      (0x41, 'A'),
      (0x42, 'B'),
      (0x43, 'C'),
      (0x44, 'D'),
      (0x45, 'E'),
      (0x46, 'F'),
      (0x47, 'G'),
      (0x48, 'H'),
      (0x49, 'I'),
      (0x4a, 'J'),
      (0x4b, 'K'),
      (0x4c, 'L'),
      (0x4d, 'M'),
      (0x4e, 'N'),
      (0x4f, 'O'),
      (0x50, 'P'),
      (0x51, 'Q'),
      (0x52, 'R'),
      (0x53, 'S'),
      (0x54, 'T'),
      (0x55, 'U'),
      (0x56, 'V'),
      (0x57, 'W'),
      (0x58, 'X'),
      (0x59, 'Y'),
      (0x5a, 'Z'),

      (0x5f, 'underscore'),

      (0x61, 'a'),
      (0x62, 'b'),
      (0x63, 'c'),
      (0x64, 'd'),
      (0x65, 'e'),
      (0x66, 'f'),
      (0x67, 'g'),
      (0x68, 'h'),
      (0x69, 'i'),
      (0x6a, 'j'),
      (0x6b, 'k'),
      (0x6c, 'l'),
      (0x6d, 'm'),
      (0x6e, 'n'),
      (0x6f, 'o'),
      (0x70, 'p'),
      (0x71, 'q'),
      (0x72, 'r'),
      (0x73, 's'),
      (0x74, 't'),
      (0x75, 'u'),
      (0x76, 'v'),
      (0x77, 'w'),
      (0x78, 'x'),
      (0x79, 'y'),
      (0x7a, 'z'),
]

def mapCharToName(c):
    return {
        '0': 'zero',
        '1': 'one',
        '2': 'two',
        '3': 'three',
        '4': 'four',
        '5': 'five',
        '6': 'six',
        '7': 'seven',
        '8': 'eight',
        '9': 'nine',
        '_': 'underscore',
    }.get(c, c)

for code, name in ligaChars:
    glyph = font.createChar(code, name)
    pen = glyph.glyphPen()
    pen.moveTo(0,0)

lookup_name = '\'liga\' Standard Ligatures in Latin lookup 0'
subtable_name = lookup_name + ' subtable'
font.addLookup( lookup_name, "gsub_ligature", [], [ ("liga",[("latn",["dflt"])]) ] )
font.addLookupSubtable(lookup_name, subtable_name)

for code, name, path, options in select:
    glyph = font.createChar(code, name)
    glyph.importOutlines(path)

    ligature_components = ' '.join( map(mapCharToName, list(name)) )
    glyph.addPosSub(subtable_name, ligature_components)
    if options:
        lt = parse_transformation(options.get('font_transformation', None)) 
        if lt:
            glyph.transform(lt)
    if global_transform:
        glyph.transform(global_transform)

    if verbose:
        print("\n%s:\n  file: %s\n  code: %s" % (name, path, hex(code)))

for format in output_formats:
    font.generate('%s/%s.%s' % (build_dir, font.fontname, format))



# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import sys
import ast
import importlib
import shutil
from pathlib import Path
from bawr.environment import Environment
from bawr import utils


def load_config_file(path):
    with open(path, 'r') as f:
        return ast.parse(f.read())


class Analyzer(ast.NodeVisitor):
    BUILDER_REGISTRY = ('Font', 'Atlas', 'CppFontHeader', 'CppEmbedded', 'ImGuiFontLoader', 'CppAtlasHeader')

    def __init__(self):
        super().__init__()
        self.iconsets = dict()
        self.builders = dict()
        self.env = None

    def visit_ClassDef(self, node):
        for base in node.bases:
            kind = base.id
            if kind in Analyzer.BUILDER_REGISTRY:
                self.builders[node.name] = (kind, None)
            elif kind == 'IconSet':
                self.iconsets[node.name] = None
            elif kind == 'Environment':
                if self.env is not None:
                    print("[Error] Only one environment definition is allowed: line {}".format(node.lineno))
                    exit(-1)
                self.env = node.name


class Parser:
    data = None
    analyzer = None
    env = None

    def __init__(self, source, args):
        self._parse(source)
        self._environment(source, args['src_dir'], args['out_dir'])
        self._clean()
        self._prepare()
        self._iconsets()
        self._builders()
        self._end()

    def _parse(self, source):
        path = Path(source)
        analyzer = Analyzer()
        analyzer.visit(load_config_file(path))
        self.data = importlib.import_module(path.stem, source)
        self.analyzer = analyzer

    def _environment(self, cfg_file, src_dir, out_dir):
        print("_" * 80)
        if self.analyzer.env:
            ctor = getattr(self.data, self.analyzer.env)
        else:
            ctor = Environment
        self.env = ctor()
        self.env.build(cfg_file, src_dir, out_dir)

    def _prepare(self):
        try:
            if not self.env.BAWR_OUTPUT_DIR.exists():
                os.mkdir(self.env.BAWR_OUTPUT_DIR)
        except:
            sys.exit("Error creating output directory: %s" % self.env.BAWR_OUTPUT_DIR)   

    def _iconsets(self):
        print("_" * 80)
        for iconset in self.analyzer.iconsets:
            ctor = getattr(self.data, iconset)
            inst = ctor()
            inst.build(self.env)
            ctor.instance = inst

    def _builders(self):
        for builder in self.analyzer.builders:
            print("_" * 80)
            ctor = getattr(self.data, builder)
            inst = ctor()
            inst.build(self.env)
            ctor.instance = inst

    def _clean(self):
        cache_dir = utils.get_cache_dir(self.env)
        witness = Path(cache_dir, "__cfg__")
        if not witness.exists():
            shutil.rmtree(cache_dir, ignore_errors=True)
        else:
            if witness.stat().st_mtime < self.env.BAWR_CONFIG_FILE.stat().st_mtime:
                shutil.rmtree(cache_dir, ignore_errors=True)

    def _end(self):
        cache_dir = utils.get_cache_dir(self.env)
        witness = Path(cache_dir, "__cfg__")
        with open(witness, 'w') as f:
            f.write(f"{self.env.BAWR_CONFIG_FILE}")